C     **********************************************************
C     This program takes the quasiparticle dispersion as input
C     **********************************************************

      PROGRAM MEM2_ELIASHBERG
      IMPLICIT NONE
      INCLUDE 'MEM_inc.f'
      INTEGER NDRAW, ND, NA
      DOUBLE PRECISION Y(NMAX), D(NMAX), SIGMA(NMAX), KERN(NMAX, NMAX)
      DOUBLE PRECISION A(NMAX), M(NMAX), Y1(NMAX), EM(NMAX, NMAX)
      DOUBLE PRECISION ERAW(NMAX), KRAW(NMAX), K(NMAX)
      CHARACTER*32 DATAIN, MODEL, OUTPRX, FILENAME
      DOUBLE PRECISION KT, ECUTOFF, A1, A2, DA1, DA2, KF, EF, E0
      INTEGER METHOD, ITERNUM, NBIN, FITBPD
      DOUBLE PRECISION OMEGAD, OMEGAM, LAMBDA0
      DOUBLE PRECISION YMIN, YMAX, DY1, OMEGABIN(NMAX), BETA
      DOUBLE PRECISION ALPHA, DA, DALPHA, ERRB0, ERRB1
      DOUBLE PRECISION LAMBDA, DLAMBDA, OMEGALOG
      DOUBLE PRECISION EBX(NMAX), EBY(NMAX), EBDX(NMAX), EBDY(NMAX),
     $     D1(NMAX)
      DOUBLE PRECISION CHI, XCHI, CHI0, Q, S, IMS, IMSIGMA
      INTEGER I, J, LOUTPRX
      DOUBLE PRECISION X1, X2, X12, S1, S2, XX


      OPEN(10, FILE = 'CONF3.INI')

      READ(10, *) DATAIN
      READ(10, *) MODEL
      READ(10, *) OUTPRX
      
C     * DATA PARAMETERS *
      READ(10, *) NDRAW
      READ(10, *) NA
      READ(10, *) ECUTOFF
      READ(10, *) KT

C     * RAW DATA CORRECTION PARAMETERS *
      READ(10, *) FITBPD
      READ(10, *) A1
      READ(10, *) A2
      READ(10, *) EF
      READ(10, *) KF

C     * ERROR BARS *
      READ(10, *) ERRB0
      READ(10, *) ERRB1

C     * FITTING PARAMETERS *
      READ(10, *) METHOD
      READ(10, *) ITERNUM
      READ(10, *) ALPHA
      READ(10, *) DALPHA
      READ(10, *) XCHI

C     * DEFAULT MODEL PARAMETERS *
      READ(10, *) OMEGAD
      READ(10, *) OMEGAM
      READ(10, *) LAMBDA0

      READ(10, *) BETA
C     * OUTPUT PARAMETERS *
      READ(10, *) NBIN
      DO I = 1, NBIN+1
         READ(10, *) OMEGABIN(I)
      END DO
      CLOSE(10)

C     ****************************************************************
      LOUTPRX = INDEX(OUTPRX, ' ')-1

      FILENAME = OUTPRX(1:LOUTPRX)//'.LOG'
      OPEN(10, FILE = FILENAME)

      WRITE(10, *) '***************************************************'
      WRITE(10, *) 'INPUT DATA         : ', DATAIN
      WRITE(10, *) 'INPUT MODEL        : ', MODEL
      WRITE(10, *) 'OUTPUT FILE PREFIX : ', OUTPRX
      
      WRITE(10, *) 'TOTAL DATA POINTS  : ', ND
      WRITE(10, *) 'TOTAL OMEGA POINTS : ', NA
      WRITE(10, *) 'CUTOFF ENERGY      : ', ECUTOFF
      WRITE(10, *) 'TEMPERATURE        : ', KT

C     * RAW DATA CORRECTION PARAMETERS *
      WRITE(10, *) 'A1                 : ', A1
      WRITE(10, *) 'A2                 : ', A2
      WRITE(10, *) 'EF                 : ', EF
      WRITE(10, *) 'KF                 : ', KF


C     * ERROR BARS *
      IF(ERRB0 .LE. 0.0D0)THEN
         WRITE(10, *) 'DATA ERROR BAR     : AUTOMATIC'
      ELSE
         WRITE(10, *) 'DATA ERROR BAR     : ', ERRB0
         WRITE(10, *) 'DATA ERROR BAR SLOP: ', ERRB1
      END IF

C     * FITTING PARAMETERS *
      IF(METHOD .EQ. 1)THEN
         WRITE(10, *) 'MEM METHOD         : HISTORIC'
         WRITE(10, *) 'TARGET CHI^2       : ', XCHI
         WRITE(10, *) 'MAX ALPHA          : ', ALPHA
      ELSE IF(METHOD .EQ. 2)THEN
         WRITE(10, *) 'MEM METHOD         : CLASSIC'
      ELSE IF(METHOD .EQ. 3)THEN
         WRITE(10, *) 'MEM METHOD         : BRYAN'
         WRITE(10, *) 'MAX ALPHA          : ', ALPHA
         WRITE(10, *) 'ALPHA STEP         : ', DALPHA
      ELSE
         WRITE(10, *) 'MEM METHOD         : NONE'
      END IF

      WRITE(10, *) 'MAX ITERATION NUMB : ', ITERNUM


C     * DEFAULT MODEL PARAMETERS *
      IF(MODEL .EQ. 'NONE')THEN
         WRITE(10, *) 'DEF MODEL OMEGAD   : ', OMEGAD
         WRITE(10, *) 'DEF MODEL OMEGAM   : ', OMEGAM
         WRITE(10, *) 'DEF MODEL LAMBDA0  : ', LAMBDA0
      END IF

      WRITE(10, *) 'WEITHT EXPONENT    : ', BETA
      WRITE(10, *) 'TOTAL OUT BINS     : ', NBIN
 200  FORMAT(8F10.2)
      WRITE(10, 200) (OMEGABIN(I), I = 1, NBIN+1)
      WRITE(10, *) '***************************************************'
      WRITE(10, *) 'RESULTS:'
      WRITE(*, *) 'RESULTS:'

C     ***************************************************

      OPEN(1, FILE=DATAIN)
      DO I = 1, NDRAW
         READ(1, *) ERAW(I), KRAW(I)
         ERAW(I) = ERAW(I) - EF
         KRAW(I) = KRAW(I) - KF
      END DO
      CLOSE(1)

      ND = 0
      DO I = 1, NDRAW
         IF(ABS(ERAW(I)) .LE. ECUTOFF .AND. ERAW(I) .LT. 0.0D0)THEN
            ND = ND + 1
            Y(ND) = ERAW(I)
            D(ND) = ERAW(I) - A1 *  KRAW(I) - A2 * KRAW(I)**2
            K(ND) = KRAW(I)
         END IF
      END DO

      WRITE(10, *) 'ND                 = ', ND
      WRITE(* , *) 'ND                 = ', ND

C     ****************************************************
      CALL POST_PROCESS(ND, Y, D, SIGMA, ERRB0, ERRB1)

      KT = KT / 1.1604D4

      CALL DSCAL(ND, 1.0D0/KT, Y, 1)
      CALL DSCAL(ND, 1.0D0/KT, D, 1)
      CALL DSCAL(ND, 1.0D0/KT, SIGMA, 1)
      
      CALL DSCAL(NBIN+1, 1D-3/KT, OMEGABIN, 1)

      OMEGAD = OMEGAD / KT
      OMEGAM = OMEGAM / KT


      IF(MODEL .NE. "NONE")THEN
         OPEN(1, FILE = MODEL)
         DO I = 1, NA
            READ(1, *) Y1(I), M(I)
            Y1(I) = Y1(I) * 1D-3 / KT
            M(I) = M(I) * LAMBDA0
         END DO
         DY1 = Y1(2) - Y1(1)
         CLOSE(1)
      ELSE
         DY1 = OMEGAM / DBLE(NA)

         DO I = 1, NA
            Y1(I) = DBLE(I) * DY1
         END DO

         DO I = 1, NA
            IF(Y1(I) .LE. OMEGAD)THEN
               M(I) = LAMBDA0 * (Y1(I) / OMEGAD)**2
            ELSE IF(Y1(I) .LE. OMEGAM)THEN
               M(I) = LAMBDA0
            ELSE
               M(I) = ZERO
            END IF
            IF(M(I) .LE. 0.0D0) M(I) = ZERO
         END DO
      END IF


      X1 = 0.0D0
      X2 = 0.0D0
      X12 = 0.0D0
      DO I = 1, ND
         X1 = X1 + K(I)**2 / SIGMA(I)**2
         X2 = X2 + K(I)**4 / SIGMA(I)**2
         X12 = X12 + K(I)**3 / SIGMA(I)**2
      END DO
      XX = X1*X2 - X12**2
      X1 = X1 / XX
      X2 = X2 / XX
      X12 = -X12 / XX

      CALL SETUP_KERN(ND, NA, Y, Y1, DY1, KERN)

C     ********************************************
      DA1 = 1.0D0
      DA2 = 1.0D0
      A1 = -A1
      A2 = -A2
      J = 0
      DO WHILE((DA1 .GT. 1D-3 .OR. DA2 .GT. 1D-2) .AND. J .LE. 200)
         J = J + 1
         IF(METHOD .EQ. 1)THEN
            CALL MEMFIT_HST(ND, NA, ITERNUM, KERN, D, SIGMA, M, 
     $           A, DA, ALPHA, XCHI, EM)
         ELSE IF(METHOD .EQ. 2)THEN
            CALL MEMFIT_CLS(ND, NA, ITERNUM, KERN, D, SIGMA, M, 
     $           A, DA, ALPHA, DALPHA, EM)
         ELSE IF(METHOD .EQ. 3)THEN
            CALL MEMFIT_BRYAN(ND, NA, ITERNUM, KERN, D, SIGMA, M, 
     $           A, DA, ALPHA, DALPHA, EM)
         ELSE IF(METHOD .EQ. 4)THEN
            CALL MEMFIT_FIXALPHA(ND, NA, ITERNUM, KERN, D, SIGMA, M, 
     $           A, DA, ALPHA, DALPHA, EM)        
         ELSE
            STOP
         END IF

         IF(FITBPD .EQ. 0)GOTO 1000
            
         
         CALL DGEMV('N', ND, NA, 1.0D0, KERN, ND, A, 1, 0.0D0, D1, 1)

         S1 = 0.0D0
         S2 = 0.0D0
         DO I = 1, ND
            S1 = S1 + KT * (Y(I)+ D1(I)) * K(I) / SIGMA(I)**2
            S2 = S2 + KT * (Y(I)+ D1(I)) * K(I)**2 / SIGMA(I)**2
         END DO

         DA1 = A1
         DA2 = A2

         A1 =  X2 * S1 + X12 * S2
         A2 =  X12 * S1 + X1 * S2
         
         DA1 = ABS(DA1 - A1) / ABS(A1)
         DA2 = ABS(DA2 - A2) / ABS(A2)

         DO I = 1, ND
            D(I) = (A1 *  K(I) + A2 * K(I)**2) / KT - Y(I)
         END DO
         
         PRINT*, J, A1, A2
      END DO

 1000 A1 = -A1
      A2 = -A2

      WRITE(*, *) 'A1                 = ', A1
      WRITE(*, *) 'A2                 = ', A2

      WRITE(10, *) 'A1                 = ', A1
      WRITE(10, *) 'A2                 = ', A2


      CHI0 = CHI(ND, NA, KERN, D, SIGMA, A)
      S = 0.0D0
      DO I = 1, NA
         S = S + A(I) - M(I) + A(I) * LOG(A(I) / M(I))
      END DO
      Q = CHI0 / 2.0D0 - ALPHA * S

      WRITE(10, *) 'CHI^2              = ', CHI0
      WRITE(10, *) 'Q                  = ', Q

      WRITE(*, *) 'CHI^2              = ', CHI0
      WRITE(*, *) 'Q                  = ', Q

      IF(METHOD .EQ. 1 .OR. METHOD .EQ. 2)THEN
         WRITE(10, *) 'ALPHA              = ', ALPHA
         WRITE(*, *) 'ALPHA              = ', ALPHA
      END IF
      IF(METHOD .EQ. 1)THEN
         WRITE(10, *) 'DALPHA             = ', DALPHA
         WRITE(*, *) 'DALPHA             = ', DALPHA
      END IF

      CALL DGEMV('N', ND, NA, 1.0D0, KERN, ND, A, 1, 0.0D0, D1, 1)

      FILENAME = OUTPRX(1:LOUTPRX)//'_DAT.dat'
      OPEN(1, FILE = FILENAME) 
      DO I = 1, ND
         IMS = IMSIGMA(NA, Y(I), A, Y1, DY1)
         WRITE(1, 300) Y(I)*KT*1D3, D(I)*KT*1D3, SIGMA(I)*KT*1D3, 
     $        D1(I)*KT*1D3, IMS*KT*1D3
      END DO
      CLOSE(1)

      FILENAME = OUTPRX(1:LOUTPRX)//'_SPT.dat'
      OPEN(1, FILE = FILENAME)
      DO I = 1, NA
         WRITE(1, 100) Y1(I)*KT*1000.0, A(I), M(I)
      END DO
      CLOSE(1)


      CALL WEIGHT(NA, NBIN, OMEGABIN, BETA, A, Y1, DY1, EM, 
     $     EBX, EBY, EBDX, EBDY)

      FILENAME = OUTPRX(1:LOUTPRX)//'_WGT.dat'
      OPEN(1, FILE = FILENAME)
      DO I = 1, NBIN
         WRITE(1, 100) EBX(I)*KT*1000.0, EBY(I), EBDX(I)*KT*1000.0,
     $        EBDY(I)
      END DO
      CLOSE(1)

      CALL INTAVG(NA, A, Y1, DY1, EM, LAMBDA, DLAMBDA, OMEGALOG)
      
      WRITE(10, *) 'LAMBDA             = ', LAMBDA
      WRITE(10, *) 'DLAMBDA            = ', DLAMBDA
      WRITE(10, *) 'OMEGALOG           = ', OMEGALOG * KT * 1D3
      
      WRITE(*, *) 'LAMBDA             = ', LAMBDA
      WRITE(*, *) 'DLAMBDA            = ', DLAMBDA
      WRITE(*, *) 'OMEGALOG           = ', OMEGALOG * KT * 1D3

      CLOSE(10)

C     ************ DISPERSION OUTPUT **************************
      FILENAME = OUTPRX(1:LOUTPRX)//'_DSP.dat'
      OPEN(1, FILE = FILENAME)

      CALL DSCAL(NDRAW, -1.0D0/KT, ERAW, 1)
      CALL SETUP_KERN(NDRAW, NA, ERAW, Y1, DY1, KERN)
      CALL DGEMV('N', NDRAW, NA, 1.0D0, KERN, NDRAW, A, 1, 0.0D0, D1, 1)
      
      DO I = 1, NDRAW
         E0 = - (ERAW(I) + D1(I)) * KT
         K(I) = 2.0D0*E0/(ABS(A1) + SQRT(A1**2+4.0D0*A2*E0)) * 
     $        DSIGN(-1.0D0, A1) 
         IMS = IMSIGMA(NA, ERAW(I), A, Y1, DY1)
         WRITE(1, 300) -ERAW(I) * KT, KRAW(I), K(I), 
     $        IMS * KT, 
     $        2.0D0 * IMS * KT /ABS(A1 + 2.0D0 * A2 * K(I))
      END DO
      CLOSE(1)


 100  FORMAT(4G20.8)
 300  FORMAT(5G16.8)


      STOP
      END


C     *********************************************************

      SUBROUTINE POST_PROCESS(ND, E, D, SIGMA, ERRB0, ERRB1)
      IMPLICIT NONE
      INCLUDE 'MEM_inc.f'
      INTEGER NBIN
      PARAMETER(NBIN = 9)
      INTEGER ND
      DOUBLE PRECISION D(ND), E(ND), SIGMA(ND), DV, EOFFSET, EF, 
     $     ERRB0, ERRB1, SIGMABAR
      DOUBLE PRECISION Q(NBIN, 3), QD(NBIN), QTQ(3, 3), QTD(3)
      DOUBLE PRECISION DNRM2
      INTEGER I, J, K, IL
      INTEGER IPIV(3), INFO


      DO I = 1, ND
         E(I) =  - E(I)
      END DO
      
      IF(ERRB0 .LE. 0D0)THEN
         DO I = 1, ND
            IL = I - NBIN/2
            IF(IL .LE. 0) IL = 1
            IF(IL + NBIN .GT. ND) IL = ND - NBIN + 1

            DO J = 1, NBIN
               QD(J) = D(IL + J - 1)
            END DO

            DO K = 1, 3
               DO J = 1, NBIN
                  Q(J, K) = E(IL + J - 1) ** (K-1)
               END DO
            END DO
         
            CALL DGEMM('T', 'N', 3, 3, NBIN, 1.0D0, Q, NBIN,
     $           Q, NBIN, 0.0D0, QTQ, 3)
            CALL DGEMV('T', NBIN, 3, 1.0D0, Q,  NBIN, QD, 1,
     $           0.0D0, QTD, 1 )

            CALL DGESV(3, 1, QTQ, 3, IPIV, QTD, 3, INFO) 


            CALL DGEMV('N', NBIN, 3, 1.0D0, Q, NBIN, QTD, 1, 
     $           -1.0D0, QD, 1)

            SIGMA(I) = DNRM2(NBIN, QD, 1) / SQRT(DBLE(NBIN-3))
         END DO

         SIGMABAR = 0.0D0
         DO I = 1, ND
            SIGMABAR = SIGMABAR + SIGMA(I)**2
         END DO
         SIGMABAR = SQRT(SIGMABAR / DBLE(ND))

         IF(ERRB0 .EQ. 0.0)THEN
            DO I = 1, ND
               SIGMA(I) = SIGMABAR
            END DO
         END IF
         
         WRITE(10, *) 'SIGMABAR           = ', SIGMABAR
         WRITE(* , *) 'SIGMABAR           = ', SIGMABAR
      ELSE
         DO I = 1, ND
            SIGMA(I) = ERRB0 + ERRB1 * E(I)
         END DO
      END IF

      RETURN
      END
      

      SUBROUTINE SETUP_KERN(ND, NA, Y, Y1, DY1, KERN)
      IMPLICIT NONE
      INCLUDE 'MEM_inc.f'
      INTEGER ND, NA
      DOUBLE PRECISION Y(ND), Y1(NA), DY1, KERN(ND, NA)
      DOUBLE PRECISION GN, G
      INTEGER I, J, N
      
      DO J = 1, NA
         DO I = 1, ND

            G = 0.0D0
            GN = 1.0D0
            N = 0
            DO WHILE (ABS(GN) .GT. 1D-6 * ABS(G))
               GN = 8 * PI2 * DBLE(2*N+1) * Y(I) * Y1(J) /
     $              ((Y(I)-Y1(J))**2+DBLE(2*N+1)**2*PI2) /
     $              ((Y(I)+Y1(J))**2+DBLE(2*N+1)**2*PI2)
               G = G + GN
               N = N + 1
            END DO
            KERN(I, J) = G * DY1

         END DO
      END DO

      RETURN
      END


C     ***********************************************
      SUBROUTINE WEIGHT(NA, NBIN, OMEGABIN, BETA, A, Y1, DY1, EM,
     $     EBX, EBY, EBDX, EBDY)
      IMPLICIT NONE
      INCLUDE 'MEM_inc.f'
      INTEGER NA, NBIN
      DOUBLE PRECISION A(NA), Y1(NA), DY1, EM(NA, NA), OMEGABIN(*), BETA
      DOUBLE PRECISION EBX(NBIN), EBY(NBIN), EBDX(NBIN), EBDY(NBIN)
      INTEGER  I, J, K, L, L0, L1, NL

      L = 1
      DO WHILE(Y1(L) .LT. OMEGABIN(1))
         L = L + 1
      END DO

      DO I = 1, NBIN
         
         EBX(I)  = (OMEGABIN(I+1) + OMEGABIN(I))/2.0D0
         EBDX(I) = (OMEGABIN(I+1) - OMEGABIN(I))/2.0D0

         L0 = L
         EBY(I) = 0.0D0
         DO WHILE(Y1(L) .LT. OMEGABIN(I+1) .AND. L .LE. NA)
            EBY(I) = EBY(I) + A(L) * Y1(L)**BETA
            L = L + 1
         END DO
         L1 = L - 1

         EBY(I) = EBY(I) * DY1

         EBDY(I) = 0.0D0
         DO J = L0, L1
            DO K = L0, L1
               EBDY(I) = EBDY(I) + EM(J, K) * (Y1(J)*Y1(K))**BETA
            END DO
         END DO
         EBDY(I) = SQRT(EBDY(I))* DY1
      END DO

      RETURN
      END




      SUBROUTINE INTAVG(NA, A, Y1, DY, EM, LAMBDA, DLAMBDA,
     $     OMEGALOG)
      IMPLICIT NONE
      INCLUDE 'MEM_inc.f'
      INTEGER NA
      DOUBLE PRECISION A(NA), Y1(NA), DY, EM(NA, NA)
      DOUBLE PRECISION LAMBDA, DLAMBDA, OMEGALOG
      INTEGER I, J

      LAMBDA = 0.0D0
      DO I = 1, NA
         LAMBDA = LAMBDA + A(I) / Y1(I)
      END DO
      LAMBDA = 2.0D0 * LAMBDA * DY
      
      DLAMBDA = 0.0D0
      DO J =1, NA
         DO I = 1, NA
            DLAMBDA = DLAMBDA + EM(I, J) / Y1(I) / Y1(J)
         END DO
      END DO

      DLAMBDA = SQRT(DLAMBDA) * 2.0D0 * DY


      OMEGALOG = 0.0D0
      DO I = 1, NA
         OMEGALOG = OMEGALOG + A(I) / Y1(I) * LOG(Y1(I))
      END DO
      OMEGALOG = EXP(2.0D0 * OMEGALOG * DY / LAMBDA)
      
      RETURN
      END
      

C     *************************************************


      DOUBLE PRECISION FUNCTION IMSIGMA(NA, Y, AF, Y1, DY1)
      IMPLICIT NONE
      INCLUDE 'MEM_inc.f'
      INTEGER NA
      DOUBLE PRECISION Y, KT, AF(NA), Y1(NA), DY1
      DOUBLE PRECISION F, NB
      INTEGER I

      IMSIGMA = 0.0D0

      DO I = 1, NA
         IMSIGMA = IMSIGMA + AF(I) * 
     $        (F(Y1(I)-Y)+F(Y1(I)+Y) + 2.0D0 * NB(Y1(I)))
      END DO

      IMSIGMA = IMSIGMA * PI * DY1
      
      RETURN
      END




      DOUBLE PRECISION FUNCTION NORM_RAND(SIGMA)
      IMPLICIT NONE
      DOUBLE PRECISION SIGMA, RAND
      INTEGER I, N

      N = 10

      NORM_RAND = 0.0D0
      DO I = 1, N
         NORM_RAND = NORM_RAND + SQRT(12.0D0) * (RAND() - 0.5D0) * SIGMA
      END DO

      NORM_RAND = NORM_RAND / DBLE(N)
      
      RETURN
      END
      

      double precision function f(x)
      double precision x

      if(x.ge. 0)then
         f = exp(-x)/(exp(-x)+1.0d0)
      else
         f = 1.0d0/(exp(x) + 1.0d0)
      end if

      return
      end

      
      double precision function nb(x)

      double precision x

      nb = exp(-x)/(-exp(-x)+1.0d0)

      return
      end
